/* -
 * Copyright (c) 2002, Ramsey G. Brenner. <rgbrenner@myrealbox.com>
 * All rights reserved.
 *
 * The contents of this file are subject to the terms of the XORSuite
 * Public Source License (the "License"); You may not use this file except
 * in compliance with the License. You may obtain a copy of the License
 * at http://rgbrenner.cjb.net/projects/xpsl.html.
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 */

static const char copyright[] =
"@(#) Copyright (c) 2002\n\
  Ramsey G. Brenner <rgbrenner@myrealbox.com>. All rights reserved.\n";

#include "hdrs.h"

void   split(FILE *, FILE **, FILE *, int);

char  *cwd;
char  *tmp_dir;
char  *prog_name;

/* Split or combine file(s) */
int
main(int argc, char **argv)
{
    tf      fflag;
    int     i;
    char   *tmp,
          **files,
           *out_name;
    size_t  size,
            nkeys,
            nfiles;

    set_general_variables();

    /* Init some variables */
    nkeys = 1;
    nfiles = 0;
    size = -1;
    fflag = FALSE;
    out_name = NULL;
    files = Malloc(sizeof(char *));
    files[0] = NULL;

    /* Init options */
    for(i = 1; i < argc; i++) {
      switch(get_opt_type(argv[i])) {
        case HASH_SHA1:      /* Set hash_algo */
          set_hash(SH);
          break;

        case HASH_MD5:      /* Set hash_algo */
          set_hash(MD);
          break;

        case BGEN_SSL:      /* Set bgen_algo */
          set_bgen(SSL);
          break;

        case BGEN_URANDOM:  /* Set bgen_algo */
          set_bgen(URANDOM);
          break;

        case N_OPT:         /* Set number of keys */
          if((tmp = get_opt(argv + i, &i)) == NULL) {
            usage();
            exit(EXIT_FAILURE);
          }
          if(strisdigit(tmp)) {
            nkeys = Atoi(tmp);
          } else {
            usage();
            exit(EXIT_FAILURE);
          }
          break;

        case OUTPUT_FILE:   /* Set output name */
          if((out_name = get_opt(argv + i, &i)) == NULL) {
            usage();
            exit(EXIT_FAILURE);
          }
          break;

        case SIZE:        /* Set size of output files */
          if((tmp = get_opt(argv + i, &i)) == NULL) {
            usage();
            exit(EXIT_FAILURE);
          }
          if(strisdigit(tmp)) {
            size = Atoi(tmp);
          } else {
            usage();
            exit(EXIT_FAILURE);
          }
          break;

        case FFLAG:         /* Set fflag */
          fflag = TRUE;
          break;

        case OTHER:         /* Add to file list */
        default:
          nfiles++;
          files = Realloc(files, (nfiles + 1) * sizeof(char *));
          files[nfiles] = NULL;
          files[nfiles - 1] = argv[i];
          break;
      }
    }

    /* Check usage */
    if(nfiles == 0) {
      usage();
      exit(EXIT_FAILURE);
    }

    out_name = set_cwd(out_name);

    /* split/combine file(s) */
    if(nfiles == 1) {  /* Split file */
      char    *in_name,
             **key_name;
      FILE    *in_fp,
              *out_fp,
             **key_fp;

      /* Init out_name */
      out_name = NULL;

      /* Assign name to in_name */
      in_name = files[0];

      /* Make sure input file exists */
      if(!chk_file(in_name)) {
        Err(MISSING_FILE, in_name);
      }

      /* Open input file */
      in_fp = Fopen(in_name, "r");

      /* Allocate space to store key file names */
      key_name = Malloc((nkeys + 1) * sizeof(char *));
      /* Allocate space to store key FILE ptrs */
      key_fp = Malloc((nkeys + 1) * sizeof(FILE *));
      /* Assign NULLs to end of space */
      key_name[nkeys] = NULL;
      key_fp[nkeys] = NULL;

      /* Open key files */
      for(i = 0; i < nkeys; i++) {
        /* Open temporary FILE */
        key_fp[i] = Tmpfile();
      }

      /* Open output file */
      /* Open temporary FILE */
      out_fp = Tmpfile();

      /* Trim input file to size */
      trim(&in_fp, size);

      /* Split file */
      split(in_fp, key_fp, out_fp, nkeys);

      /* Copy keys to a tmp fp */
      for(i = 0; i < nkeys; i++) {
        key_fp[i] = Copy(key_fp[i], &key_name[i]);
      }

      /* Copy out_fp to a tmp fp */
      out_fp = Copy(out_fp, &out_name);

      /* Close outputed files */
      /* Close input file */
      Fclose(in_fp);
      /* Close key files */
      for(i = 0; i < nkeys; i++) {
        Fclose(key_fp[i]);
      }
      /* Close output file */
      Fclose(out_fp);

      /* Rename key files */
      for(i = 0; i < nkeys; i++) {
        /* Rename key */
        Rename(key_name[i], NULL, "key", fflag);
      }

      /* Rename output file */
      Rename(out_name, NULL, "dat", fflag);

      /* Cleanup */
      Free(key_fp);
      Free(key_name);
      Free(out_name);
    } else {  /* Combine files */
      int      n_infp; /* Number of input files */
      char   **in_name,
              *tmp_name;
      FILE   **in_fp,
              *out_fp;

      /* Init n_infp */
      n_infp = nfiles;

      /* Allocate space for input names */
      in_name = Malloc((n_infp + 1) * sizeof(char *));
      /* Allocate space to store FILE ptrs */
      in_fp = Malloc((n_infp + 1) * sizeof(FILE *));
      /* Assign NULL to end of memory */
      in_name[n_infp] = NULL;
      in_fp[n_infp] = NULL;

      /* Assign filenames to *in_name and
         open input files */
      for(i = 0; i < n_infp; i++) {
        /* Assign name from *files to *in_name */
        in_name[i] = files[i];

        /* Make sure the file exists */
        if(!chk_file(in_name[i])) {
          Err(MISSING_FILE, in_name[i]);
        }

        /* Check the filenames hash */
        chk_hash(in_name[i]);

        /* Open input file */
        in_fp[i] = Fopen(in_name[i], "r");
      }

      /* Open temporary FILE */
      out_fp = Tmpfile();

      /* XOR files */
      XORFiles(in_fp, out_fp);

      /* Trim output file to size */
      trim(&out_fp, size);

      /* Copy output in out_fp to correct output file */
      if(out_name == NULL) {
        /* Copy to stdout */
        Copy(out_fp, NULL);
      } else {
        out_fp = Copy(out_fp, &tmp_name);
      }

      /* Close all files */
      for(i = 0; in_fp[i] != NULL; i++) {
        Fclose(in_fp[i]);
      }
      Fclose(out_fp);

      /* -o <name> was specified */
      if(out_name != NULL) {
        Rename(tmp_name, out_name, NULL, fflag);
        Free(tmp_name);
      }

      /* Cleanup */
      Free(in_fp);
      Free(in_name);
    }

    exit(EXIT_SUCCESS);
}

void
split(FILE *in_fp, FILE **key_fp, FILE *out_fp, int nkeys)
{
    int     i,
            nfp,  /* number of input files */
            length;
    FILE  **input_fp;

    /* Set length */
    length = get_length(in_fp);

    /* Generate keys */
    for(i = 0; i < nkeys; i++) {
      genkey(key_fp[i], length);
      rewind(key_fp[i]);
    }

    /* nfp = nkeys + 1 for in_fp */
    nfp = nkeys + 1;

    /* Allocate enough memory to hold all the input FILE ptrs */
    /* nkeys + 1 for in_fp + 1 for NULL */
    input_fp = Malloc((nfp + 1) * sizeof(FILE *));
    /* Assign NULL to end of memory */
    input_fp[nfp] = NULL;

    /* group in_fp and key_fp */
    input_fp[0] = in_fp;
    for(i = 1; i < nfp; i++) {
      input_fp[i] = key_fp[i - 1];
    }

    /* XOR files */
    XORFiles(input_fp, out_fp);

    /* cleanup and return */
    Free(input_fp);
    return;
}

void
usage(void)
{
    fprintf(stderr, "XORSuite v%s\n"
      "Usage: %s [options] <input files>\n"
      "Options:\n"
        "\t-f               - Do not prompt if file already exists\n"
        "\t-n <number>      - Number of keys to mix in (default: 1)\n"
        "\t-o <name>        - Output name of combined files (default: stdout)\n"
        "\t-s <size>        - Output file will be <size> bytes\n"
        "\t-sha1            - Set hash type to SHA-1 (default)\n"
        "\t-md5             - Set hash type to MD5\n"
        "\t-ssl             - Set byte generator to use SSL (default)\n"
        "\t-urandom         - Set byte generator to use /dev/urandom\n"
        "\n",
      VERSION,
      prog_name);
    return;
}
