/* -
 * Copyright (c) 2002, Ramsey G. Brenner. <rgbrenner@myrealbox.com>
 * All rights reserved.
 *
 * The contents of this file are subject to the terms of the XORSuite
 * Public Source License (the "License"); You may not use this file except
 * in compliance with the License. You may obtain a copy of the License
 * at http://rgbrenner.cjb.net/projects/xpsl.html.
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 */

extern char *cwd;
extern char *tmp_dir;
extern char *prog_name;

F_REG       *tfp;

void
chk_cwd(void)
{
    struct stat  sb;

    if(*cwd == NULL) {
      return;
    }

    if(stat(cwd, &sb) < 0) {
      if(errno == ENOENT) {
        Errx("%s: Path does not exist",
          cwd);
      } else {
        if(errno != EIO && errno != EFAULT) {
          Errx("%s: %s",
            cwd, strerror(errno));
        } else {
          perror(NULL);
          exit(EXIT_FAILURE);
        }
      }
    }

    if(!S_ISDIR(sb.st_mode)) {
      Errx("%s: Not a directory",
        cwd);
    }

    return;
}

void
set_tmpdir(void)
{
    tmp_dir = getenv("TMPDIR");
    if(tmp_dir == NULL) {
      tmp_dir = "/tmp/";
    }
}

/* chdir() */
void
Chdir(const char *path)
{
    if(chdir(path) < 0) {
      perror(NULL);
      exit(EXIT_FAILURE);
    }
    return;
}

/* Replacement for the non-standard mkstemp() function */
/* Starting from the end of tmplt, it replaces all the 'X's
   with random chars until it encouters a non-'X' */
int
Mkstemp(char **tmplt)
{
    int            i,
                   c,
                   d,
                   nchrs;
    char          *tptr,
                  *tname,
                  *chrs;
    size_t         tlen;
    unsigned char  byte;

    tptr = *tmplt;

    tlen = strlen(tptr);
    tname = Malloc(tlen + 1);

    /* nlowercase + nuppercase + nnumbers */
    nchrs = 26 + 26 + 10;

    chrs = Malloc(nchrs + 1);

    memset(chrs, 0, nchrs + 1);

    i = 0;
    for(c = 'a'; i < 26; i++, c++) {
      chrs[i] = c;
    }

    for(c = '0'; i < 10 + 26; i++, c++) {
      chrs[i] = c;
    }

    for(c = 'A'; i < 26 + 10 + 26; i++, c++) {
      chrs[i] = c;
    }

    for(;;) {
      /* Replace X's with */
      memset(tname, 0, tlen + 1);
      for(i = tlen - 1; i >= 0; i--) {
        if(tptr[i] == 'X') {
          if(genbyte(&byte)) {
            Err(BYTE_GEN, NULL);
          }

          c = byte % nchrs;
          tname[i] = chrs[c];
        } else {
          for(; i >= 0; i--) {
            tname[i] = tptr[i];
          }
          break;
        }
      }

      /* Attempt open */
      if((d = open(tname, O_CREAT|O_EXCL|O_RDWR, S_IRUSR|S_IWUSR)) < 0) {
        /* If failed because file already exists */
        if(errno == EEXIST) {
          errno = 0;
          continue;
        } else {
          if(errno != EIO && errno != EFAULT) {
            Errx("%s: %s",
              tname, strerror(errno));
          } else {
            perror(NULL);
            exit(EXIT_FAILURE);
          }
        }
      }

      break;
    }

    /* Copy tname to tmplt */
    for(i = 0; i <= tlen; i++) {
      tptr[i] = tname[i];
    }

    Free(chrs);
    Free(tname);
    return(d);
}

/* Unlink all file pointers */
void
del_tfp(void)
{
    while(tfp != NULL) {
      Fclose(tfp->fp);
    }
}

/* fclose() */
/* If fp was opened using Tmpfile(),
   the file name will be removed */
void
Fclose(FILE *fp)
{
    F_REG   *pre,
            *ptr;

    /* Init */
    pre = NULL;
    ptr = tfp;

    while(ptr != NULL) {
      if(ptr->fp == fp) {

        /* Remove entry from tfp */
        if(pre == NULL) {
          tfp = ptr->next;
        } else {
          pre->next = ptr->next;
        }

        fclose(ptr->fp);
        Unlink(ptr->fn);
        Free(ptr->fn);
        Free(ptr);
        return;
      }
      pre = ptr;
      ptr = ptr->next;
    }

    /* fp was not a temp file */
    fclose(fp);

    return;
}

/* Replacement for the insecure tmpfile() function */
FILE *
Tmpfile(void)
{
    F_REG       *itm,
                *last;
    static tf    func_set = FALSE;

    /* Find the last entry in tfp */
    last = tfp;
    while(last != NULL) {
      if(last->next != NULL) {
        last = last->next;
      } else {
        break;
      }
    }

    /* Allocate mem for new itm */
    itm = Malloc(sizeof(F_REG));

    /* Open temp file */
    itm->fp = tmp_fp(tmp_dir, &itm->fn);

    /* Set itm->next */
    itm->next = NULL;

    /* Append itm to end of tfp */
    if(tfp != NULL) {
      last->next = itm;
    } else {
      tfp = itm;
    }

    if(func_set == FALSE) {
      /* Setup atexit() to unlink all files */
      AtExit(del_tfp);
      func_set = TRUE;
    }

    /* Return tmp FILE ptr */
    return(itm->fp);
}

/* Check if fn is a directory */
tf
isdir(const char *p)
{
    struct stat  sb;

    if(p == NULL) {
      return(FALSE);
    }

    if(*p == NULL) {
      return(FALSE);
    }

    if(stat(p, &sb) < 0) {
      if(errno == ENOENT) {
        errno = 0;
        return(FALSE);
      } else {
        if(errno != EIO && errno != EFAULT) {
          Errx("%s: %s",
            p, strerror(errno));
        } else {
          perror(NULL);
          exit(EXIT_FAILURE);
        }
      }
    }

    if(S_ISDIR(sb.st_mode)) {
      return(TRUE);
    } else {
      return(FALSE);
    }
}

/* set cwd */
char *
set_cwd(const char *fn)
{
    char  *path,
          *filename;

    if(fn == NULL) {
      cwd = Malloc(2);
      memset(cwd, 0, 2);
      return(NULL);
    } else {
      cwd = path = get_path(fn);
      filename = path + strlen(path) + 1;

      chk_cwd();

      if(*filename == NULL) {
        return(NULL);
      } else {
        return(filename);
      }
    }
}

/* return path */
char *
get_path(const char *fn)
{
    int    i;
    char  *p,
          *path,
          *filename;

    if(isdir(fn)) {
      if(fn[strlen(fn) - 1] == '/') {
        path = Malloc(strlen(fn) + 2);
        sprintf(path, "%s", fn);
        path[strlen(fn) + 1] = NULL;
      } else {
        path = Malloc(strlen(fn) + 3);
        sprintf(path, "%s/", fn);
        path[strlen(fn) + 2] = NULL;
      }
      return(path);
    } else {
      Asprintf(&path, "");
      Asprintf(&filename, "%s", fn);
    }

    for(i = strlen(fn); i > 0; i--) {
      if(fn[i - 1] == '/') {
        Free(path);
        Free(filename);
        path = Malloc(i + 1);
        strncpy(path, fn, i);
        path[i] = NULL;
        Asprintf(&filename, "%s", fn + i);
        break;
      }
    }

    p = Malloc((strlen(path) + 1) + (strlen(filename) + 1));

    strcpy(p, path);
    strcpy(p + strlen(path) + 1, filename);

    Free(path);
    Free(filename);

    return(p);
}

/* read one byte from stream */
int
Rbyte(void *ptr, FILE *stream)
{
    /* Fill bytes */
    if(!fread(ptr, sizeof(unsigned char), 1, stream)) {
      if(feof(stream)) {
        return(1);
      } else {
        Err(READ, NULL);
      }
    }

    return(0);
}

/* write one byte to stream */
void
Wbyte(const void *ptr, FILE *stream)
{
    if(!fwrite(ptr, sizeof(unsigned char), 1, stream)) {
      Err(WRITE, NULL);
    }
}

/* Rename tmp_name */
void
Rename(const char *tmp_name, const char *out_name,
       const char *postfix, tf fflag)
{
    char  *outfn;

    if(out_name == NULL) {
      outfn = hash_rename(tmp_name, postfix);
      /* Tell user about new file */
      printf("Wrote %s\n",
        outfn);
    } else {
      Asprintf(&outfn, "%s%s",
        cwd, out_name);

      /* Make sure we aren't about to overwrite one
         of the users other files */
      if(chk_file(outfn)) {
        if(fflag == FALSE) {
          /* We found a file named the same as our output file */
          /* Ask user if we should overwrite */
          Warn("Output file `%s' already exists",
            outfn);
          switch(ask_user("Overwrite? [y,N] ")) {
            case 'y':
              Unlink(outfn);
              break;
            case 'n':
            case '\n':
              /* Remove tmp file */
              Unlink(tmp_name);
              return;
          }
        } else {
          Unlink(outfn);
        }
      }

      /* Rename tmp_name to out_name */
      if(rename(tmp_name, outfn) < 0) {
        Err(RENAME, tmp_name);
        out_name = tmp_name;
      }

      /* Give user hash of new file */
      printf("%s Hash of %s: %s\n",
        get_hash_algo(),
        outfn, hash_file(outfn));

      Free(outfn);
    }
    return;
}

/* Create a temporary file and return a FILE ptr */
FILE *
tmp_fp(char *dir, char **name)
{
    tf       istmp;
    int      fd;
    char    *d,
            *tmp;
    FILE    *fp;
    size_t   dlen;

    /* Init variables */
    d = dir;
    istmp = FALSE;
    if(d == NULL) {
      dlen = 0;
    } else {
      dlen = strlen(d);
    }

    /* If d is not a NULL string */
    if(dlen > 0) {
      if(d[dlen - 1] != '/') {
        /* set istmp */
        istmp = TRUE;

        /* Allocate memory for d + '/' + NULL */
        tmp = Malloc(dlen + 2);

        /* Copy d to tmp and append '/' */
        strcpy(tmp, d);
        d[dlen] = '/';
        d[dlen + 1] = NULL;

        /* strlen(d) + strlen("/") */
        dlen++;

        /* Set d to tmp */
        d = tmp;
      }
    }

    /* prepend dir to tmp name */
    Asprintf(name, "%stmp.XXXXXXXX",
      d);

    /* Open unique tmp name */
    fd = Mkstemp(name);

    /* Open FILE ptr */
    if((fp = fdopen(fd, "r+")) == NULL) {
      close(fd);
      Unlink(*name);
      Errx("%s: %s",
        *name, strerror(errno));
    }

    /* If istmp is set, free tmp */
    if(istmp == TRUE) {
      Free(tmp);
    }

    /* Return new FILE ptr */
    return(fp);
}

/* Check if file exists */
tf
chk_file(const char *fn)
{
    struct stat sb;

    if(stat(fn, &sb) < 0) {
      if(errno == ENOENT) {
        errno = 0;
        return(FALSE);
      } else {
        if(errno != EIO && errno != EFAULT) {
          Errx("%s: %s",
            fn, strerror(errno));
        } else {
          perror(NULL);
          exit(EXIT_FAILURE);
        }
      }
    }
    return(TRUE);
}

/* Return file length */
size_t
get_length(FILE *fp)
{
    struct stat sb;

    if(fstat(fileno(fp), &sb) < 0) {
      perror(NULL);
      exit(EXIT_FAILURE);
    }
    return(sb.st_size);
}

/* fopen() */
FILE *
Fopen(const char *path, const char *mode)
{
    FILE  *tmp;

    if((tmp = fopen(path, mode)) == NULL) {
      if(errno != EIO && errno != EFAULT) {
        Errx("%s: %s",
          path, strerror(errno));
      } else {
        perror(NULL);
        exit(EXIT_FAILURE);
      }
    }
    return(tmp);
}

/* unlink() */
void
Unlink(const char *path)
{
    if(unlink(path) < 0) {
      Warn("Could not remove `%s'",
        path);
    }
    return;
}
